import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestDemag:
    def setup(self):
        p1 = (-10e-9, 0, 0)
        p2 = (10e-9, 1e-9, 1e-9)
        cell = (1e-9, 1e-9, 1e-9)
        region = df.Region(p1=p1, p2=p2)
        self.mesh = df.Mesh(region=region, cell=cell)

    def test_demag(self):
        name = 'demag'

        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Demag()

        system.m = df.Field(self.mesh, dim=3, value=(1, 1, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        # Check if it runs. Tests to be added here.

    def test_demag_asymptotic_radius(self):
        name = 'demag_asymptotic_radius'

        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Demag(asymptotic_radius=6)

        system.m = df.Field(self.mesh, dim=3, value=(0, 0, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        # Check if it runs. Tests to be added here.



import oommfc as oc


def test_version():
    assert isinstance(oc.__version__, str)
    assert '.' in oc.__version__


def test_dependencies():
    assert isinstance(oc.__dependencies__, list)
    assert len(oc.__dependencies__) > 0



from .driver import Driver


class MinDriver(Driver):
    """Energy minimisation driver.

    Only attributes in ``_allowed_attributes`` can be defined. For details on
    possible values for individual attributes and their default values, please
    refer to ``Oxs_MinDriver`` documentation (https://math.nist.gov/oommf/).

    Examples
    --------
    1. Defining driver with a keyword argument.

    >>> import oommfc as oc
    ...
    >>> md = oc.MinDriver(stopping_mxHxm=0.01)

    2. Passing an argument which is not allowed.

    >>> import oommfc as oc
    ...
    >>> md = oc.MinDriver(myarg=1)
    Traceback (most recent call last):
       ...
    AttributeError: ...

    3. Getting the list of allowed attributes.

    >>> import oommfc as oc
    ...
    >>> md = oc.MinDriver()
    >>> md._allowed_attributes
    [...]

    """
    _allowed_attributes = ['evolver',
                           'stopping_mxHxm',
                           'stage_iteration_limit',
                           'total_iteration_limit',
                           'stage_count',
                           'stage_count_check',
                           'checkpoint_file',
                           'checkpoint_interval',
                           'checkpoint_disposal',
                           'start_iteration',
                           'start_stage',
                           'start_stage_iteration',
                           'start_stage_start_time',
                           'start_stage_elapsed_time',
                           'start_last_timestep',
                           'normalize_aveM_output',
                           'report_max_spin_angle',
                           'report_wall_time']

    def _checkargs(self, **kwargs):
        pass  # no kwargs should be checked



import numbers
import numpy as np
import discretisedfield as df


def box_atlas(pmin, pmax, name):
    mif = f'# BoxAtlas for {name}_atlas\n'
    mif += f'Specify Oxs_BoxAtlas:{name}_atlas {{\n'
    mif += f'  xrange {{ {pmin[0]} {pmax[0]} }}\n'
    mif += f'  yrange {{ {pmin[1]} {pmax[1]} }}\n'
    mif += f'  zrange {{ {pmin[2]} {pmax[2]} }}\n'
    mif += f'  name {name}\n'
    mif += '}\n\n'

    return mif


def atlas_vector_field(value, name, atlas='main_atlas'):
    mif = f'# {name}\n'
    mif += f'Specify Oxs_AtlasVectorField:{name} {{\n'
    mif += f'  atlas :{atlas}\n'
    mif += '  default_value {{{} {} {}}}\n'.format(*value['default'])
    mif += '  values {\n'
    for region, val in value.items():
        if region != 'default':
            mif += '    {} {{{} {} {}}}\n'.format(region, *val)
    mif += '  }'
    mif += '}\n\n'

    return mif


def atlas_scalar_field(value, name, atlas='main_atlas'):
    mif = f'# {name}\n'
    mif += f'Specify Oxs_AtlasScalarField:{name} {{\n'
    mif += f'  atlas :{atlas}\n'
    mif += '  default_value {}\n'.format(value['default'])
    mif += '  values {\n'
    for region, val in value.items():
        if region != 'default':
            mif += f'    {region} {val}\n'
    mif += '  }'
    mif += '}\n\n'

    return mif


def setup_m0(field, name):
    field.write(f'{name}.omf')
    # Magnetisation
    mif = file_vector_field(f'{name}.omf', f'{name}', 'main_atlas')
    # Ms
    mif += vector_norm_scalar_field(f'{name}', f'{name}_norm')

    return mif, f'{name}', f'{name}_norm'


def file_vector_field(filename, name, atlas):
    mif = f'# {name} file\n'
    mif += f'Specify Oxs_FileVectorField:{name} {{\n'
    mif += f'  file {filename}\n'
    mif += f'  atlas :{atlas}\n'
    mif += '}\n\n'

    return mif


def vector_norm_scalar_field(field, name):
    mif = f'# {name}\n'
    mif += f'Specify Oxs_VecMagScalarField:{name} {{\n'
    mif += f'    field :{field}\n'
    mif += '}\n\n'

    return mif


def setup_scalar_parameter(parameter, name):
    if isinstance(parameter, df.Field):
        parameter.write(f'{name}.ovf', extend_scalar=True)
        mif = file_vector_field(f'{name}.ovf', f'{name}', 'main_atlas')
        mif += vector_norm_scalar_field(f'{name}', f'{name}_norm')
        return mif, f'{name}_norm'

    elif isinstance(parameter, dict):
        if 'default' not in parameter.keys():
            parameter['default'] = 0
        mif = atlas_scalar_field(parameter, f'{name}')
        return mif, f'{name}'

    elif isinstance(parameter, numbers.Real):
        return '', f'{parameter}'


def setup_vector_parameter(parameter, name):
    if isinstance(parameter, df.Field):
        parameter.write(f'{name}.ovf')
        mif = file_vector_field(f'{name}.ovf', f'{name}', 'main_atlas')
        return mif, f'{name}'

    elif isinstance(parameter, dict):
        if 'default' not in parameter.keys():
            parameter['default'] = (0, 0, 0)
        mif = atlas_vector_field(parameter, f'{name}')
        return mif, f'{name}'

    elif isinstance(parameter, (tuple, list, np.ndarray)):
        return '', '{{{} {} {}}}'.format(*parameter)



import oommfc as oc
import micromagneticmodel as mm


def driver_script(driver, system, compute=None, **kwargs):
    mif = ''
    if isinstance(driver, oc.MinDriver):
        # Check evolver and set default if not passed.
        if not hasattr(driver, 'evolver'):
            driver.evolver = oc.CGEvolver()
        elif not isinstance(driver.evolver, oc.CGEvolver):
            msg = f'Cannot use {type(driver.evolver)} for evolver.'
            raise TypeError(msg)

        # Define default stopping_mxHxm if not passed. OOMMF cannot run without
        # this value.
        if not hasattr(driver, 'stopping_mxHxm'):
            driver.stopping_mxHxm = 0.1

        mif += oc.scripts.evolver_script(driver.evolver)

        # Minimisation driver script.
        mif += '# MinDriver\n'
        mif += 'Specify Oxs_MinDriver {\n'
        mif += '  evolver :evolver\n'
        mif += '  mesh :mesh\n'
        mif += '  Ms :m0_norm\n'
        mif += '  m0 :m0\n'
        for attr, value in driver:
            if attr != 'evolver':
                mif += f'  {attr} {value}\n'
        mif += '}\n\n'

        # Saving results.
        mif += 'Destination table mmArchive\n'
        mif += 'Destination mags mmArchive\n\n'
        mif += 'Schedule DataTable table Stage 1\n'
        mif += 'Schedule Oxs_MinDriver::Magnetization mags Stage 1'

    if isinstance(driver, oc.TimeDriver):
        # Check evolver and set default if not passed.
        if not hasattr(driver, 'evolver'):
            if mm.ZhangLi() in system.dynamics:
                driver.evolver = oc.SpinTEvolver()
            elif mm.Slonczewski() in system.dynamics:
                driver.evolver = oc.SpinXferEvolver()
            else:
                driver.evolver = oc.RungeKuttaEvolver()
        elif not isinstance(driver.evolver, (oc.EulerEvolver,
                                             oc.RungeKuttaEvolver,
                                             oc.SpinTEvolver,
                                             oc.SpinXferEvolver)):
            msg = f'Cannot use {type(driver.evolver)} for evolver.'
            raise TypeError(msg)

        # Extract dynamics equation parameters.
        if mm.Precession() in system.dynamics:
            driver.evolver.gamma_G = system.dynamics.precession.gamma0
        else:
            driver.evolver.do_precess = 0
        if mm.Damping() in system.dynamics:
            driver.evolver.alpha = system.dynamics.damping.alpha
        else:
            driver.evolver.alpha = 0
        if mm.ZhangLi() in system.dynamics:
            driver.evolver.u = system.dynamics.zhangli.u
            driver.evolver.beta = system.dynamics.zhangli.beta
        if mm.Slonczewski() in system.dynamics:
            driver.evolver.J = system.dynamics.slonczewski.J
            driver.evolver.mp = system.dynamics.slonczewski.mp
            driver.evolver.P = system.dynamics.slonczewski.P
            driver.evolver.Lambda = system.dynamics.slonczewski.Lambda
            driver.evolver.eps_prime = system.dynamics.slonczewski.eps_prime

        mif += oc.scripts.evolver_script(driver.evolver)

        # Extract time and number of steps.
        t, n = kwargs['t'], kwargs['n']

        # TimeDriver
        mif += '# TimeDriver\n'
        mif += 'Specify Oxs_TimeDriver {\n'
        mif += '  evolver :evolver\n'
        mif += '  mesh :mesh\n'
        mif += f'  Ms :m0_norm\n'
        mif += f'  m0 :m0\n'
        mif += f'  stopping_time {t/n}\n'
        mif += f'  stage_count {n}\n'
        for attr, value in driver:
            if attr != 'evolver':
                mif += f'  {attr} {value}\n'
        mif += '}\n\n'

        # Saving results
        mif += 'Destination table mmArchive\n'
        mif += 'Destination mags mmArchive\n'
        mif += 'Destination archive mmArchive\n\n'
        mif += 'Schedule DataTable table Stage 1\n'
        mif += 'Schedule Oxs_TimeDriver::Magnetization mags Stage 1\n'

        if compute is not None:
            mif += compute

    return mif



import pytest
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestMagnetoElastic:
    def setup(self):
        p1 = (-7e-9, -5e-9, -4e-9)
        p2 = (7e-9, 5e-9, 4e-9)
        self.region = df.Region(p1=p1, p2=p2)
        self.cell = (1e-9, 1e-9, 1e-9)
        self.subregions = {'r1': df.Region(p1=(-7e-9, -5e-9, -4e-9),
                                           p2=(0, 5e-9, 4e-9)),
                           'r2': df.Region(p1=(0, -5e-9, -4e-9),
                                           p2=(7e-9, 5e-9, 4e-9))}

    @pytest.mark.travis
    def test_scalar_scalar_vector_vector(self):
        name = 'magnetoelastic_scalar_scalar_vector_vector'

        B1 = 1e5
        B2 = 1e6
        e_diag = (1, 1, 1)
        e_offdiag = (0, 0, 0)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.MagnetoElastic(B1=B1, B2=B2, e_diag=e_diag,
                                          e_offdiag=e_offdiag)

        mesh = df.Mesh(region=self.region, cell=self.cell)
        system.m = df.Field(mesh, dim=3, value=(0, 0.3, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        # Assertions have to be invented. Checks only if it runs.



from .cgevolver import CGEvolver
from .eulerevolver import EulerEvolver
from .rungekuttaevolver import RungeKuttaEvolver
from .spintevolver import SpinTEvolver
from .spinxferevolver import SpinXferEvolver



import pytest
import pkg_resources
import oommfc.oommf
import oommfc.scripts
from .delete import delete
from .compute import compute
from .drivers import MinDriver, TimeDriver
from .evolvers import CGEvolver, EulerEvolver, RungeKuttaEvolver, \
    SpinTEvolver, SpinXferEvolver


def test():
    return pytest.main(['-m', 'not travis and not docker',
                        '-v', '--pyargs', 'oommfc'])  # pragma: no cover


def test_docker():
    return pytest.main(['-m', 'docker', '-v',
                        '--pyargs', 'oommfc'])  # pragma: no cover


__version__ = pkg_resources.get_distribution(__name__).version
__dependencies__ = pkg_resources.require(__name__)



import os
import abc
import sys
import time
import datetime
import logging
import shutil
import oommfc as oc
import subprocess as sp
import ubermagutil as uu
import micromagneticmodel as mm

log = logging.getLogger(__name__)
_cached_oommf_runner = None


class OOMMFRunner(metaclass=abc.ABCMeta):
    """Abstract class for running OOMMF.

    """
    def call(self, argstr, need_stderr=False):
        """Calls OOMMF by passing ``argstr`` to OOMMF.

        Parameters
        ----------
        argstr : str

            Argument string passed to OOMMF.

        need_stderr : bool

            If ``need_stderr=True``, standard error is captured. Defaults to
            ``False``.

        Raises
        ------
        RuntimeError

            If an error occured.

        Returns
        -------
        int

            When the OOMMF run was successful, ``0`` is returned.

        Examples
        --------
        1. Getting OOMMF runner automatically and calling it.

        >>> import oommfc as oc
        ...
        >>> runner = oc.oommf.get_oommf_runner()
        >>> runner.call(argstr='+version')
        Running OOMMF...
        CompletedProcess(...)

        """
        now = datetime.datetime.now()
        timestamp = '{}/{:02d}/{:02d} {:02d}:{:02d}'.format(now.year,
                                                            now.month,
                                                            now.day,
                                                            now.hour,
                                                            now.minute)
        print(f'Running OOMMF ({self.__class__.__name__}) [{timestamp}]... ',
              end='')

        tic = time.time()
        res = self._call(argstr=argstr, need_stderr=need_stderr)
        self._kill()  # kill OOMMF (mostly needed on Windows)
        toc = time.time()
        seconds = '({:0.1f} s)'.format(toc - tic)
        print(seconds)  # append seconds to the previous print.

        if res.returncode != 0:
            if sys.platform != 'win32':
                # Only on Linux and MacOS - on Windows we do not get stderr and
                # stdout.
                stderr = res.stderr.decode('utf-8', 'replace')
                stdout = res.stdout.decode('utf-8', 'replace')
                cmdstr = ' '.join(res.args)
                print('OOMMF error:')
                print(f'\tcommand: {cmdstr}')
                print(f'\tstdout: {cmdstr}')
                print(f'\tstderr: {stderr}')
                print('\n')
            raise RuntimeError('Error in OOMMF run.')

        return res

    @abc.abstractmethod
    def _call(self, argstr, need_stderr=False):
        """This method should be implemented in subclass.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _kill(self, targets=('all',)):
        """This method should be implemented in subclass.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def errors(self):
        """Returns the content of ``boxsii.errors`` OOMMF file.

        Returns
        -------
        str

            ``boxsii.errors`` OOMMF file.

        """
        pass  # pragma: no cover

    def version(self):
        """Returns the OOMMF version.

        Returns
        -------
        str

            OOMMF version.

        Examples
        --------
        1. Getting OOMMF version.

        >>> import oommfc as oc
        ...
        >>> runner = oc.oommf.get_oommf_runner()
        >>> runner.version()
        Running OOMMF...
        '...'

        """
        res = self.call(argstr='+version', need_stderr=True)
        return res.stderr.decode('utf-8').split('oommf.tcl')[-1].strip()

    def platform(self):
        """Returns platform seen by OOMMF.

        Returns
        -------
        str

            Platform.

        Examples
        --------
        1. Getting platform.

        >>> import oommfc as oc
        ...
        >>> runner = oc.oommf.get_oommf_runner()
        >>> runner.platform()
        Running OOMMF...
        '...'

        """
        res = self.call(argstr='+platform', need_stderr=True)
        return res.stderr.decode('utf-8')


@uu.inherit_docs
class TclOOMMFRunner(OOMMFRunner):
    """OOMMF runner using path to ``oommf.tcl``.

    Parameters
    ----------
    oommf_tcl: str

        Path to ``oommf.tcl``file.

    """
    def __init__(self, oommf_tcl):
        self.oommf_tcl = oommf_tcl  # a path to oommf.tcl

    def _call(self, argstr, need_stderr=False):
        cmd = ['tclsh', self.oommf_tcl, 'boxsi', '+fg',
               argstr, '-exitondone', '1']

        # Not clear why we cannot get stderr and stdout on win32. Calls to
        # OOMMF get stuck.
        stdout = stderr = sp.PIPE
        if sys.platform == 'win32' and not need_stderr:
            stdout = stderr = None  # pragma: no cover

        return sp.run(cmd, stdout=stdout, stderr=stderr)

    def _kill(self, targets=['all']):
        sp.run(['tclsh', self.oommf_tcl, 'killoommf'] + targets)

    def errors(self):
        errors_file = os.path.join(os.path.dirname(self.oommf_tcl),
                                   'boxsi.errors')
        with open(errors_file, 'r') as f:
            errors = f.read()

        return errors


@uu.inherit_docs
class ExeOOMMFRunner(OOMMFRunner):
    """OOMMF runner using OOMMF executable, which can be found on $PATH.

    Parameters
    ----------
    oommf_exe: str

        Name of the OOMMF executable. Defaults to ``oommf``.

    """
    def __init__(self, oommf_exe='oommf'):
        self.oommf_exe = oommf_exe

    def _call(self, argstr, need_stderr=False):
        # Here we might need stderr = stdot = None like in
        # TclOOMMFRunner for Windows.  This is not clear because we
        # never use ExeOOMMFRunner on Windows.
        cmd = [self.oommf_exe, 'boxsi', '+fg', argstr, '-exitondone', '1']
        return sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)

    def _kill(self, targets=['all']):
        sp.run([self.oommf_exe, 'killoommf'] + targets)

    def errors(self):
        try:
            errors_file = os.path.join(
                os.path.dirname(shutil.which(self.oommf_exe)),
                '..', 'opt', 'oommf', 'boxsi.errors')
            with open(errors_file, 'r') as f:
                errors = f.read()
            return errors

        except FileNotFoundError:
            msg = 'boxsi.errors cannot be retrieved.'
            raise EnvironmentError(msg)


@uu.inherit_docs
class DockerOOMMFRunner(OOMMFRunner):
    """OOMMF runner using Docker.

    Parameters
    ----------
    docker_exe: str

        Docker executable. Defaults to ``docker``.

    image: str

        Docker image on DockerHub. Defaults to ``ubermag/oommf``.

    """
    def __init__(self, docker_exe='docker', image='ubermag/oommf'):
        self.docker_exe = docker_exe
        self.image = image

    def _call(self, argstr, need_stderr=False):
        cmd = [self.docker_exe, 'run', '-v', os.getcwd()+':/io',
               self.image, '/bin/bash', '-c',
               ('tclsh /usr/local/oommf/oommf/oommf.tcl boxsi '
                '+fg {} -exitondone 1').format(argstr)]
        return sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)

    def _kill(self, targets=('all',)):
        # There is no need to kill OOMMF when run inside docker.
        pass

    def errors(self):
        msg = 'boxsi.errors cannot be retrieved from Docker container.'
        raise EnvironmentError(msg)


def get_oommf_runner(use_cache=True, envvar='OOMMFTCL',
                     oommf_exe='oommf', docker_exe='docker'):
    """Find the best available way to run OOMMF.

    Returns an ``oommfc.oommf.OOMMFRunner`` object, or raises
    ``EnvironmentError`` if no suitable method is found.

    Parameters
    ----------
    use_cache : bool

      The first call to this function will determine the best way to run OOMMF
      and cache it. Normally, subsequent calls will return the ``OOMMFRunner``
      object from the cache. Setting this parameter to ``False`` will cause it
      to check for available methods again. Defaults to ``True``.

    envvar : str

      Name of the environment variable containing the path to ``oommf.tcl``.
      Defaults to ``'OOMMFTCL'``.

    oommf_exe : str

      The name or path of the executable ``oommf`` command. Defaults to
      ``'oommf'``.

    docker_exe : str

      The name or path of the docker command. Defaults to ``'docker'``.

    Returns
    -------
    oommfc.oommf.OOMMFRunner

        An OOMMF runner.

    Raises
    ------
    EnvironmentError

        If no OOMMF can be found on host.

    Examples
    --------
    1. Getting OOMMF Runner.

    >>> import oommfc as oc
    ...
    >>> runner = oc.oommf.get_oommf_runner()
    >>> isinstance(runner, oc.oommf.OOMMFRunner)
    True

    """
    global _cached_oommf_runner
    if use_cache and (_cached_oommf_runner is not None):
        return _cached_oommf_runner

    # Check for the OOMMFTCL environment variable pointing to oommf.tcl.
    oommf_tcl = os.environ.get(envvar, None)
    if oommf_tcl is not None:
        cmd = ['tclsh', oommf_tcl, 'boxsi',
               '+fg', '+version', '-exitondone', '1']
        try:
            res = sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
        except FileNotFoundError:
            log.warning('oommf.tcl was not found.')
        else:
            if res.returncode:
                log.warning('OOMMFTCL is set, but OOMMF could not be run.\n'
                            f'stdout:\n{res.stdout}\n'
                            f'stderr:\n{res.stderr}')
            else:
                _cached_oommf_runner = TclOOMMFRunner(oommf_tcl)
                return _cached_oommf_runner

    # OOMMF is installed via conda and oommf.tcl is in opt/oommf (Windows).
    # This would probably also work on MacOS/Linux, but on these operating
    # systems, when installed via conda, we use 'oommf' executable.
    if sys.platform == 'win32' and \
       os.path.isdir(os.path.join(sys.prefix, 'conda-meta')):
        oommf_tcl = os.path.join(sys.prefix, 'opt', 'oommf', 'oommf.tcl')
        if os.path.isfile(oommf_tcl):
            _cached_oommf_runner = TclOOMMFRunner(oommf_tcl)
            return _cached_oommf_runner

    # OOMMF available as an executable - in a conda env on Mac/Linux, or oommf
    # installed separately.
    oommf_exe = shutil.which(oommf_exe)
    if oommf_exe:
        _cached_oommf_runner = ExeOOMMFRunner(oommf_exe)
        return _cached_oommf_runner

    # Check for docker to run OOMMF in a docker image.
    cmd = [docker_exe, 'images']
    try:
        res = sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    except FileNotFoundError:
        log.warning('Docker was not found.')
    else:
        if res.returncode:
            log.warning('Error running docker\n'
                        f'stdout:\n{res.stdout}\n'
                        f'stderr:\n{res.stderr}')
        else:
            _cached_oommf_runner = DockerOOMMFRunner(docker_exe=docker_exe,
                                                     image='ubermag/oommf')
            return _cached_oommf_runner

    # If OOMMFRunner was not returned up to this point, we raise an
    # exception.
    raise EnvironmentError('Cannot find OOMMF.')


def status():
    """Run a macrospin example for 1 ps through oommfc and print the OOMMF
    status.

    Returns
    -------
    int

        If ``0``, the OOMMF is found and running. Otherwise, ``1`` is returned.

    Examples
    --------
    1. Checking the OOMMF status.

    >>> import oommfc as oc
    ...
    >>> oc.oommf.status()
    Running OOMMF...
    OOMMF found and running.
    0

    """
    system = mm.examples.macrospin()
    try:
        td = oc.TimeDriver()
        td.drive(system, t=1e-12, n=1)
        print('OOMMF found and running.')
        return 0
    except (EnvironmentError, RuntimeError):
        print('Cannot find OOMMF.')
        return 1


def overhead():
    """Run a macrospin example for 1 ps through ``oommfc`` and directly and
    return the difference in run times.

    Returns
    -------
    float

      The time difference (overhead) between running OOMMF though ``oommfc``
      and directly.

    Examples
    --------
    1. Getting the overhead time.

    >>> import oommfc as oc
    ...
    >>> isinstance(oc.oommf.overhead(), float)
    Running OOMMF...
    True

    """
    # Running OOMMF through oommfc.
    system = mm.examples.macrospin()
    td = oc.TimeDriver()
    oommfc_start = time.time()
    td.drive(system, t=1e-12, n=1, save=True, overwrite=True)
    oommfc_stop = time.time()
    oommfc_time = oommfc_stop - oommfc_start

    # Running OOMMF directly.
    oommf_runner = get_oommf_runner()
    mifpath = os.path.realpath(os.path.join(system.name,
                                            'drive-0',
                                            'macrospin.mif'))
    oommf_start = time.time()
    oommf_runner.call(mifpath)
    oommf_stop = time.time()
    oommf_time = oommf_stop - oommf_start
    oc.delete(system)

    return oommfc_time - oommf_time



import numpy as np
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestZeeman:
    def setup(self):
        p1 = (-10e-9, -5e-9, -3e-9)
        p2 = (10e-9, 5e-9, 3e-9)
        self.region = df.Region(p1=p1, p2=p2)
        self.cell = (1e-9, 1e-9, 1e-9)
        self.subregions = {'r1': df.Region(p1=(-10e-9, -5e-9, -3e-9),
                                           p2=(10e-9, 0, 3e-9)),
                           'r2': df.Region(p1=(-10e-9, 0, -3e-9),
                                           p2=(10e-9, 5e-9, 3e-9))}

    def test_vector(self):
        name = 'zeeman_vector'

        H = (0, 0, 1e6)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)

        mesh = df.Mesh(region=self.region, cell=self.cell)
        system.m = df.Field(mesh, dim=3, value=(1, 1, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m(mesh.region.random_point())
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

    def test_dict(self):
        name = 'zeeman_dict'

        H = {'r1': (1e5, 0, 0), 'r2': (0, 0, 1e5)}
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)

        mesh = df.Mesh(region=self.region, cell=self.cell,
                       subregions=self.subregions)
        system.m = df.Field(mesh, dim=3, value=(0, 1, 0), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-2e-9, -2e-9, -2e-9))
        assert np.linalg.norm(np.subtract(value, (Ms, 0, 0))) < 1e-3

        value = system.m((2e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

    def test_field(self):
        name = 'zeeman_field'

        def value_fun(pos):
            x, y, z = pos
            if x <= 0:
                return (1e6, 0, 0)
            else:
                return (0, 0, 1e6)

        mesh = df.Mesh(region=self.region, cell=self.cell)

        H = df.Field(mesh, dim=3, value=value_fun)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)
        system.m = df.Field(mesh, dim=3, value=(0, 1, 0), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-2e-9, -2e-9, -2e-9))
        assert np.linalg.norm(np.subtract(value, (Ms, 0, 0))) < 1e-3

        value = system.m((2e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3



import oommfc as oc


def mesh_script(mesh):
    mif = ''
    # Define atlas.
    if mesh.subregions:
        # The mesh is composed of subregions. Multiple BoxAtlas scripts are
        # generated and the main MultiAtlas.
        for name, subregion in mesh.subregions.items():
            mif += oc.scripts.box_atlas(subregion.pmin, subregion.pmax,
                                        name=name)
        mif += '# MultiAtlas\n'
        mif += 'Specify Oxs_MultiAtlas:main_atlas {\n'
        for name in mesh.subregions.keys():
            mif += f'  atlas :{name}_atlas\n'
        mif += f'  xrange {{ {mesh.region.pmin[0]} {mesh.region.pmax[0]} }}\n'
        mif += f'  yrange {{ {mesh.region.pmin[1]} {mesh.region.pmax[1]} }}\n'
        mif += f'  zrange {{ {mesh.region.pmin[2]} {mesh.region.pmax[2]} }}\n'
        mif += '}\n\n'
    else:
        # There are no subregions in the mesh and only a single BoxAtlas is
        # generated.
        mif += oc.scripts.box_atlas(mesh.region.pmin, mesh.region.pmax,
                                    name='main')

    # Define mesh.
    if mesh.pbc:
        mif += '# PeriodicRectangularMesh\n'
        mif += 'Specify Oxs_PeriodicRectangularMesh:mesh {\n'
        mif += '  cellsize {{ {} {} {} }}\n'.format(*mesh.cell)
        mif += '  atlas :main_atlas\n'
        mif += '  periodic {}\n'.format(''.join(sorted(mesh.pbc)))
        mif += '}\n\n'
    else:
        mif += '# RectangularMesh\n'
        mif += 'Specify Oxs_RectangularMesh:mesh {\n'
        mif += '  cellsize {{ {} {} {} }}\n'.format(*mesh.cell)
        mif += '  atlas :main_atlas\n'
        mif += '}\n\n'

    return mif



import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


def test_stdprob5():
    name = 'stdprob5'

    # Geometry
    lx = 100e-9  # x dimension of the sample(m)
    ly = 100e-9  # y dimension of the sample (m)
    lz = 10e-9  # sample thickness (m)

    # Material (permalloy) parameters
    Ms = 8e5  # saturation magnetisation (A/m)
    A = 1.3e-11  # exchange energy constant (J/m)

    # Dynamics (LLG + ZhangLi equation) parameters
    gamma0 = 2.211e5  # gyromagnetic ratio (m/As)
    alpha = 0.1  # Gilbert damping
    ux = -72.35  # velocity in x direction
    beta = 0.05  # non-adiabatic STT parameter

    system = mm.System(name=name)
    p1 = (0, 0, 0)
    p2 = (lx, ly, lz)
    cell = (5e-9, 5e-9, 5e-9)
    region = df.Region(p1=p1, p2=p2)
    mesh = df.Mesh(region=region, cell=cell)
    system.energy = mm.Exchange(A=A) + mm.Demag()

    def m_vortex(pos):
        x, y, z = pos[0]/1e-9-50, pos[1]/1e-9-50, pos[2]/1e-9
        return (-y, x, 10)

    system.m = df.Field(mesh, dim=3, value=m_vortex, norm=Ms)

    md = oc.MinDriver()
    md.drive(system)

    system.dynamics += (mm.Precession(gamma0=gamma0) +
                        mm.Damping(alpha=alpha) + mm.ZhangLi(u=ux, beta=beta))

    td = oc.TimeDriver()
    td.drive(system, t=8e-9, n=100)

    mx = system.table['mx'].values

    assert -0.35 < mx.min() < -0.30
    assert -0.03 < mx.max() < 0



import numpy as np
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestZhangLi:
    def setup(self):
        p1 = (-5e-9, -5e-9, -3e-9)
        p2 = (5e-9, 5e-9, 3e-9)
        self.region = df.Region(p1=p1, p2=p2)
        self.n = (5, 5, 5)
        self.subregions = {'r1': df.Region(p1=(-5e-9, -5e-9, -3e-9),
                                           p2=(5e-9, 0, 3e-9)),
                           'r2': df.Region(p1=(-5e-9, 0, -3e-9),
                                           p2=(5e-9, 5e-9, 3e-9))}

    def test_scalar_scalar(self):
        name = 'zhangli_scalar_scalar'

        u = 0
        beta = 0.5
        H = (0, 0, 1e5)
        Ms = 1e6

        mesh = df.Mesh(region=self.region, n=self.n)

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)
        system.dynamics = mm.ZhangLi(u=u, beta=beta)
        system.m = df.Field(mesh, dim=3, value=(0, 0.1, 1), norm=Ms)

        td = oc.TimeDriver()
        td.drive(system, t=0.2e-9, n=50)

        # u is zero, nothing should change.
        value = system.m(mesh.region.random_point())
        assert np.linalg.norm(np.cross(value, (0, 0.1*Ms, Ms))) < 1e-3

    def test_dict_scalar(self):
        name = 'zhangli_dict_scalar'

        H = (0, 0, 1e6)
        u = {'r1': 0, 'r2': 1}
        beta = 0.5
        Ms = 1e6

        mesh = df.Mesh(region=self.region, n=self.n,
                       subregions=self.subregions)

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)
        system.dynamics = mm.ZhangLi(u=u, beta=beta)
        system.m = df.Field(mesh, dim=3, value=(0, 0.1, 1), norm=Ms)

        td = oc.TimeDriver()
        td.drive(system, t=0.2e-9, n=50)

        # u=0 region
        value = system.m((1e-9, -4e-9, 3e-9))
        assert np.linalg.norm(np.cross(value, (0, 0.1*Ms, Ms))) < 1e-3

        # u!=0 region
        value = system.m((1e-9, 4e-9, 3e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) > 1

    def test_field_scalar(self):
        name = 'zhangli_field_scalar'

        mesh = df.Mesh(region=self.region, n=self.n)

        def u_fun(pos):
            x, y, z = pos
            if y <= 0:
                return 0
            else:
                return 1

        H = (0, 0, 1e6)
        u = df.Field(mesh, dim=1, value=u_fun)
        beta = 0.5
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)
        system.dynamics = mm.ZhangLi(u=u, beta=beta)
        system.m = df.Field(mesh, dim=3, value=(0, 0.1, 1), norm=Ms)

        td = oc.TimeDriver()
        td.drive(system, t=0.2e-9, n=50)

        # u=0 region
        value = system.m((1e-9, -4e-9, 3e-9))
        assert np.linalg.norm(np.cross(value, (0, 0.1*Ms, Ms))) < 1e-3

        # u!=0 region
        value = system.m((1e-9, 4e-9, 3e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) > 1



import numpy as np
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestCubicAnisotropy:
    def setup(self):
        p1 = (-7e-9, 0, 0)
        p2 = (7e-9, 5e-9, 4e-9)
        self.region = df.Region(p1=p1, p2=p2)
        self.cell = (1e-9, 1e-9, 2e-9)
        self.subregions = {'r1': df.Region(p1=(-7e-9, 0, 0),
                                           p2=(0, 5e-9, 4e-9)),
                           'r2': df.Region(p1=(0, 0, 0),
                                           p2=(7e-9, 5e-9, 4e-9))}

    def test_scalar_vector_vector(self):
        name = 'cubicanisotropy_scalar_vector_vector'

        mesh = df.Mesh(region=self.region, cell=self.cell)

        K = 1e5
        u1 = (0, 0, 1)
        u2 = (0, 1, 0)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.CubicAnisotropy(K=K, u1=u1, u2=u2)

        def m_fun(pos):
            x, y, z = pos
            if x <= 0:
                return (0, 0.2, 1)
            else:
                return (0, 1, 0.2)

        system.m = df.Field(mesh, dim=3, value=m_fun, norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-1e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

        value = system.m((1e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, Ms, 0))) < 1e-3

    def test_field_vector_vector(self):
        name = 'cubicanisotropy_field_vector_vector'

        mesh = df.Mesh(region=self.region, cell=self.cell)

        def K_fun(pos):
            x, y, z = pos
            if x <= 0:
                return 0
            else:
                return 1e5

        K = df.Field(mesh, dim=1, value=K_fun)
        u1 = (0, 0, 1)
        u2 = (0, 1, 0)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.CubicAnisotropy(K=K, u1=u1, u2=u2)
        system.m = df.Field(mesh, dim=3, value=(0, 0.3, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-2e-9, 1e-9, 1e-9))
        assert np.linalg.norm(np.cross(value, (0, 0.3*Ms, Ms))) < 1e-3

        value = system.m((2e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

    def test_field_field_field(self):
        name = 'cubicanisotropy_field_field_field'

        mesh = df.Mesh(region=self.region, cell=self.cell)

        def K_fun(pos):
            x, y, z = pos
            if -2e-9 <= x <= 2e-9:
                return 0
            else:
                return 1e5

        def u1_fun(pos):
            x, y, z = pos
            if x <= 0:
                return (0, 1, 0)
            else:
                return (0, 0, 1)

        def u2_fun(pos):
            x, y, z = pos
            if x <= 0:
                return (0, 0, 1)
            else:
                return (0, 1, 0)

        K = df.Field(mesh, dim=1, value=K_fun)
        u1 = df.Field(mesh, dim=3, value=u1_fun)
        u2 = df.Field(mesh, dim=3, value=u2_fun)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.CubicAnisotropy(K=K, u1=u1, u2=u2)
        system.m = df.Field(mesh, dim=3, value=(0, 0.3, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((0, 0, 0))
        assert np.linalg.norm(np.cross(value, (0, 0.3*Ms, Ms))) < 1e-3

        value = system.m((3e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

        value = system.m((-3e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

    def test_dict_vector_vector(self):
        name = 'cubicanisotropy_dict_vector_vector'

        mesh = df.Mesh(region=self.region, cell=self.cell,
                       subregions=self.subregions)

        K = {'r1': 0, 'r2': 1e5}
        u1 = (0, 0, 1)
        u2 = (0, 1, 0)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.CubicAnisotropy(K=K, u1=u1, u2=u2)
        system.m = df.Field(mesh, dim=3, value=(0, 0.3, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-2e-9, 1e-9, 1e-9))
        assert np.linalg.norm(np.cross(value, (0, 0.3*Ms, Ms))) < 1e-3

        value = system.m((2e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3



import random
import numpy as np
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestExchange:
    def setup(self):
        p1 = (-5e-9, -5e-9, -3e-9)
        p2 = (5e-9, 5e-9, 3e-9)
        self.region = df.Region(p1=p1, p2=p2)
        self.n = (10, 10, 6)
        self.subregions = {'r1': df.Region(p1=(-5e-9, -5e-9, -3e-9),
                                           p2=(5e-9, 0, 3e-9)),
                           'r2': df.Region(p1=(-5e-9, 0, -3e-9),
                                           p2=(5e-9, 5e-9, 3e-9))}

    def random_m(self, pos):
        return [2*random.random()-1 for i in range(3)]

    def test_scalar(self):
        name = 'exchange_scalar'

        A = 1e-12
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Exchange(A=A)

        mesh = df.Mesh(region=self.region, n=self.n)
        system.m = df.Field(mesh, dim=3, value=self.random_m, norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        assert abs(np.linalg.norm(system.m.average) - Ms) < 1e-3

    def test_dict(self):
        name = 'exchange_dict'

        A = {'r1': 0, 'r2': 1e-12, 'r1:r2': 1e-12, 'default': 2e-12}
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Exchange(A=A)

        mesh = df.Mesh(region=self.region, n=self.n,
                       subregions=self.subregions)
        system.m = df.Field(mesh, dim=3, value=self.random_m, norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        # A=0 region
        value1 = system.m((1e-9, -4e-9, 2e-9))
        value2 = system.m((1e-9, -2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value1, value2)) > 1

        # A!=0 region
        value1 = system.m((1e-9, 4e-9, 2e-9))
        value2 = system.m((1e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value1, value2)) < 1

    def test_field(self):
        name = 'exchange_field'

        def A_fun(pos):
            x, y, z = pos
            if x <= 0:
                return 1e-10  # for 0, OOMMF gives nan
            else:
                return 1e-12

        mesh = df.Mesh(region=self.region, n=self.n)
        A = df.Field(mesh, dim=1, value=A_fun)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Exchange(A=A)

        system.m = df.Field(mesh, dim=3, value=self.random_m, norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        assert abs(np.linalg.norm(system.m.average) - Ms) < 1e-3



import numpy as np
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestSlonczewski:
    def setup(self):
        p1 = (-5e-9, -5e-9, -3e-9)
        p2 = (5e-9, 5e-9, 3e-9)
        self.region = df.Region(p1=p1, p2=p2)
        self.n = (2, 2, 2)
        self.subregions = {'r1': df.Region(p1=(-5e-9, -5e-9, -3e-9),
                                           p2=(5e-9, 0, 3e-9)),
                           'r2': df.Region(p1=(-5e-9, 0, -3e-9),
                                           p2=(5e-9, 5e-9, 3e-9))}

    def test_single_values(self):
        name = 'slonczewski_scalar_values'

        J = 1e12
        mp = (1, 0, 0)
        P = 0.4
        Lambda = 2
        eps_prime = 0
        H = (0, 0, 1e6)
        Ms = 1e6

        mesh = df.Mesh(region=self.region, n=self.n)

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)
        system.dynamics = mm.Slonczewski(J=J, mp=mp, P=P, Lambda=Lambda,
                                         eps_prime=eps_prime)
        system.m = df.Field(mesh, dim=3, value=(0, 0.1, 1), norm=Ms)

        td = oc.TimeDriver()
        td.drive(system, t=0.2e-9, n=20)

        # Check if it runs.

    def test_dict_values(self):
        name = 'slonczewski_scalar_values'

        J = {'r1': 1e12, 'r2': 5e12}
        mp = {'r1': (0, 0, 1), 'r2': (0, 1, 0)}
        P = {'r1': 0.4, 'r2': 0.35}
        Lambda = {'r1': 2, 'r2': 1.5}
        eps_prime = {'r1': 0, 'r2': 1}
        H = (0, 0, 1e6)
        Ms = 1e6

        mesh = df.Mesh(region=self.region, n=self.n,
                       subregions=self.subregions)

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)
        system.dynamics = mm.Slonczewski(J=J, mp=mp, P=P, Lambda=Lambda,
                                         eps_prime=eps_prime)
        system.m = df.Field(mesh, dim=3, value=(0, 0.1, 1), norm=Ms)

        td = oc.TimeDriver()
        td.drive(system, t=0.2e-9, n=20)

        # Check if it runs.

    def test_field_values(self):
        name = 'slonczewski_scalar_values'

        mesh = df.Mesh(region=self.region, n=self.n)

        J = df.Field(mesh, dim=1, value=0.5e12)
        mp = df.Field(mesh, dim=3, value=(1, 0, 0))
        P = df.Field(mesh, dim=1, value=0.5)
        Lambda = df.Field(mesh, dim=1, value=2)
        eps_prime = df.Field(mesh, dim=1, value=1)
        H = (0, 0, 1e6)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.Zeeman(H=H)
        system.dynamics = mm.Slonczewski(J=J, mp=mp, P=P, Lambda=Lambda,
                                         eps_prime=eps_prime)
        system.m = df.Field(mesh, dim=3, value=(0, 0.1, 1), norm=Ms)

        td = oc.TimeDriver()
        td.drive(system, t=0.2e-9, n=20)

        # Check if it runs.



import oommfc as oc


def evolver_script(evolver):
    mif = ''

    # Prepare parameters depending on what attributes are defined in evolver.
    if hasattr(evolver, 'gamma_G'):
        gamma0mif, gamma0name = oc.scripts.setup_scalar_parameter(
            evolver.gamma_G, 'pr_gamma0')
        evolver.gamma_G = gamma0name
        mif += gamma0mif
    if hasattr(evolver, 'alpha'):
        alphamif, alphaname = oc.scripts.setup_scalar_parameter(
            evolver.alpha, 'dp_alpha')
        evolver.alpha = alphaname
        mif += alphamif
    if hasattr(evolver, 'u'):
        umif, uname = oc.scripts.setup_scalar_parameter(evolver.u, 'zl_u')
        evolver.u = uname
        mif += umif

    # zhangli.beta cannot spatially vary - it has to be a constant.

    if hasattr(evolver, 'J'):
        Jmif, Jname = oc.scripts.setup_scalar_parameter(evolver.J, 'sl_J')
        evolver.J = Jname
        mif += Jmif
    if hasattr(evolver, 'mp'):
        mpmif, mpname = oc.scripts.setup_vector_parameter(evolver.mp, 'sl_mp')
        evolver.mp = mpname
        mif += mpmif
    if hasattr(evolver, 'P'):
        Pmif, Pname = oc.scripts.setup_scalar_parameter(evolver.P, 'sl_P')
        evolver.P = Pname
        mif += Pmif
    if hasattr(evolver, 'Lambda'):
        Lambdamif, Lambdaname = oc.scripts.setup_scalar_parameter(
            evolver.Lambda, 'sl_Lambda')
        evolver.Lambda = Lambdaname
        mif += Lambdamif
    if hasattr(evolver, 'eps_prime'):
        eps_primemif, eps_primename = oc.scripts.setup_scalar_parameter(
            evolver.eps_prime, 'sl_eps_prime')
        evolver.eps_prime = eps_primename
        mif += eps_primemif

    # Scripts for a specific evolver.
    if isinstance(evolver, oc.EulerEvolver):
        mif += '# EulerEvolver\n'
        mif += 'Specify Oxs_EulerEvolve:evolver {\n'
    elif isinstance(evolver, oc.RungeKuttaEvolver):
        mif += '# RungeKuttaEvolver\n'
        mif += 'Specify Oxs_RungeKuttaEvolve:evolver {\n'
    elif isinstance(evolver, oc.SpinTEvolver):
        mif += '# Zhang-Li evolver\n'
        mif += 'Specify Anv_SpinTEvolve:evolver {\n'
    elif isinstance(evolver, oc.SpinXferEvolver):
        mif += '# Slonczewski evolver\n'
        mif += 'Specify Oxs_SpinXferEvolve:evolver {\n'
    elif isinstance(evolver, oc.CGEvolver):
        mif += '# CGEvolver\n'
        mif += 'Specify Oxs_CGEvolve:evolver {\n'

    # Define all other parameters.
    for attr, value in evolver:
        mif += f'  {attr} {value}\n'
    mif += '}\n\n'

    return mif



from .driver import Driver


class TimeDriver(Driver):
    """Time driver.

    Only attributes in ``_allowed_attributes`` can be defined. For details on
    possible values for individual attributes and their default values, please
    refer to ``Oxs_TimeDriver`` documentation (https://math.nist.gov/oommf/).

    Examples
    --------
    1. Defining driver with a keyword argument.

    >>> import oommfc as oc
    ...
    >>> td = oc.TimeDriver(total_iteration_limit=5)

    2. Passing an argument which is not allowed.

    >>> import oommfc as oc
    ...
    >>> td = oc.TimeDriver(myarg=1)
    Traceback (most recent call last):
       ...
    AttributeError: ...

    3. Getting the list of allowed attributes.

    >>> import oommfc as oc
    ...
    >>> td = oc.TimeDriver()
    >>> td._allowed_attributes
    [...]

    """
    _allowed_attributes = ['evolver',
                           'stopping_dm_dt',
                           'stage_iteration_limit',
                           'total_iteration_limit',
                           'stage_count_check',
                           'checkpoint_file',
                           'checkpoint_interval',
                           'checkpoint_disposal',
                           'start_iteration',
                           'start_stage',
                           'start_stage_iteration',
                           'start_stage_start_time',
                           'start_stage_elapsed_time',
                           'start_last_timestep',
                           'normalize_aveM_output',
                           'report_max_spin_angle',
                           'report_wall_time']

    def _checkargs(self, **kwargs):
        t, n = kwargs['t'], kwargs['n']
        if t <= 0:
            msg = f'Cannot drive with t={t}.'
            raise ValueError(msg)
        elif n <= 0 or not isinstance(n, int):
            msg = f'Cannot drive with n={n}.'
            raise ValueError(msg)



import re
import os
import glob
import shutil
import oommfc as oc
import ubermagtable as ut
import discretisedfield as df
import micromagneticmodel as mm


def oxs_class(term):
    """Extract the OOMMF ``Oxs_`` class name of an individual term.

    """
    mif = getattr(oc.scripts.energy, f'{term.name}_script')(term)
    return re.search(r'Oxs_([\w_]+)', mif).group(1)


def schedule_script(func):
    """Generate OOMMF ``Schedule...`` line for saving an individual value.

    """
    if func.__name__ == 'energy':
        return ''  # Datatable with energies is saved by default.
    elif func.__name__ == 'effective_field':
        if isinstance(func.__self__, mm.Energy):
            output = 'Oxs_RungeKuttaEvolve:evolver:Total field'
        else:
            output = f'Oxs_{oxs_class(func.__self__)}::Field'
    elif func.__name__ == 'density':
        if isinstance(func.__self__, mm.Energy):
            output = 'Oxs_RungeKuttaEvolve:evolver:Total energy density'
        else:
            output = f'Oxs_{oxs_class(func.__self__)}::Energy density'
    else:
        msg = f'Computing the value of {func} is not supported.'
        raise ValueError(msg)

    return 'Schedule \"{}\" archive Step 1\n'.format(output)


def compute(func, system):
    """Computes a particular value of an energy term or energy container
    (``energy``, ``density``, or ``effective_field``).

    Parameters
    ----------
    func : callable

        A property of an energy term or an energy container.

    system : micromagneticmodel.System

        Micromagnetic system for which the property is calculated.

    Returns
    -------
    numbers.Real, discretisedfield.Field

        Resulting value.

    Examples
    --------
    1. Computing values of energy terms.

    >>> import micromagneticmodel as mm
    >>> import oommfc as oc
    ...
    >>> system = mm.examples.macrospin()
    >>> oc.compute(system.energy.zeeman.energy, system)
    Running OOMMF...
    -8.8...e-22
    >>> oc.compute(system.energy.effective_field, system)
    Running OOMMF...
    Field(...)
    >>> oc.compute(system.energy.density, system)
    Running OOMMF...
    Field(...)

    """
    td = oc.TimeDriver(total_iteration_limit=1)
    td.drive(system, t=1e-25, n=1, save=True, overwrite=True,
             compute=schedule_script(func))

    if func.__name__ == 'energy':
        extension = '*.odt'
    elif func.__name__ == 'effective_field':
        extension = '*.ohf'
    elif func.__name__ == 'density':
        extension = '*.oef'

    dirname = os.path.join(system.name, f'compute-{system.drive_number}')
    output_file = max(glob.iglob(os.path.join(dirname, extension)),
                      key=os.path.getctime)

    if func.__name__ == 'energy':
        table = ut.read(output_file, rename=False)
        if isinstance(func.__self__, mm.Energy):
            output = table['RungeKuttaEvolve:evolver:Total energy'][0]
        else:
            output = table[f'{oxs_class(func.__self__)}::Energy'][0]
    else:
        output = df.Field.fromfile(output_file)

    # Delete "compute" directory after the data is extracted.
    shutil.rmtree(dirname)

    # Delete the parent directory if it remains empty after deleting "compute"
    # directory.
    if not os.listdir(system.name):
        shutil.rmtree(system.name)

    return output



import numpy as np
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestUniaxialAnisotropy:
    def setup(self):
        p1 = (-7e-9, -5e-9, -4e-9)
        p2 = (7e-9, 5e-9, 4e-9)
        self.region = df.Region(p1=p1, p2=p2)
        self.cell = (1e-9, 1e-9, 1e-9)
        self.subregions = {'r1': df.Region(p1=(-7e-9, -5e-9, -4e-9),
                                           p2=(0, 5e-9, 4e-9)),
                           'r2': df.Region(p1=(0, -5e-9, -4e-9),
                                           p2=(7e-9, 5e-9, 4e-9))}

    def test_scalar_vector(self):
        name = 'uniaxialanisotropy_scalar_vector'

        K = 1e5
        u = (0, 0, 1)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.UniaxialAnisotropy(K=K, u=u)

        mesh = df.Mesh(region=self.region, cell=self.cell)
        system.m = df.Field(mesh, dim=3, value=(0, 0.3, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m(mesh.region.random_point())
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

    def test_field_vector(self):
        name = 'uniaxialanisotropy_field_vector'

        def value_fun(pos):
            x, y, z = pos
            if x <= 0:
                return 0
            else:
                return 1e5

        mesh = df.Mesh(region=self.region, cell=self.cell)

        K = df.Field(mesh, dim=1, value=value_fun)
        u = (0, 0, 1)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.UniaxialAnisotropy(K=K, u=u)
        system.m = df.Field(mesh, dim=3, value=(0, 0.3, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-2e-9, -2e-9, -2e-9))
        assert np.linalg.norm(np.cross(value, (0, 0.3*Ms, Ms))) < 1e-3

        value = system.m((2e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

    def test_scalar_field(self):
        name = 'uniaxialanisotropy_scalar_field'

        def value_fun(pos):
            x, y, z = pos
            if x <= 0:
                return (1, 0, 0)
            else:
                return (0, 1, 0)

        mesh = df.Mesh(region=self.region, cell=self.cell)

        K = 1e5
        u = df.Field(mesh, dim=3, value=value_fun)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.UniaxialAnisotropy(K=K, u=u)
        system.m = df.Field(mesh, dim=3, value=(1, 1, 0), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-2e-9, -2e-9, -2e-9))
        assert np.linalg.norm(np.subtract(value, (Ms, 0, 0))) < 1e-3

        value = system.m((2e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, Ms, 0))) < 1e-3

    def test_field_field(self):
        name = 'uniaxialanisotropy_field_field'

        def K_fun(pos):
            x, y, z = pos
            if -2e-9 <= x <= 2e-9:
                return 0
            else:
                return 1e5

        def u_fun(pos):
            x, y, z = pos
            if x <= 0:
                return (1, 0, 0)
            else:
                return (0, 1, 0)

        mesh = df.Mesh(region=self.region, cell=self.cell)

        K = df.Field(mesh, dim=1, value=K_fun)
        u = df.Field(mesh, dim=3, value=u_fun)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.UniaxialAnisotropy(K=K, u=u)
        system.m = df.Field(mesh, dim=3, value=(1, 1, 0), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-3e-9, -3e-9, -3e-9))
        assert np.linalg.norm(np.subtract(value, (Ms, 0, 0))) < 1e-3

        value = system.m((3e-9, 3e-9, 3e-9))
        assert np.linalg.norm(np.subtract(value, (0, Ms, 0))) < 1e-3

        value = system.m((0, 0, 0))
        assert np.linalg.norm(np.cross(value, (Ms, Ms, 0))) < 1e-3

    def test_dict_vector(self):
        name = 'uniaxialanisotropy_dict_vector'

        mesh = df.Mesh(region=self.region, cell=self.cell,
                       subregions=self.subregions)
        K = {'r1': 0, 'r2': 1e5}
        u = (0, 0, 1)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.UniaxialAnisotropy(K=K, u=u)
        system.m = df.Field(mesh, dim=3, value=(0, 0.3, 1), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-2e-9, -2e-9, -2e-9))
        assert np.linalg.norm(np.cross(value, (0, 0.3*Ms, Ms))) < 1e-3

        value = system.m((2e-9, 2e-9, 2e-9))
        assert np.linalg.norm(np.subtract(value, (0, 0, Ms))) < 1e-3

    def test_field_dict(self):
        name = 'uniaxialanisotropy_field_dict'

        def K_fun(pos):
            x, y, z = pos
            if -2e-9 <= x <= 2e-9:
                return 0
            else:
                return 1e5

        def u_fun(pos):
            x, y, z = pos
            if x <= 0:
                return (1, 0, 0)
            else:
                return (0, 1, 0)

        mesh = df.Mesh(region=self.region, cell=self.cell)

        K = df.Field(mesh, dim=1, value=K_fun)
        u = df.Field(mesh, dim=3, value=u_fun)
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.UniaxialAnisotropy(K=K, u=u)
        system.m = df.Field(mesh, dim=3, value=(1, 1, 0), norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        value = system.m((-3e-9, -3e-9, -3e-9))
        assert np.linalg.norm(np.subtract(value, (Ms, 0, 0))) < 1e-3

        value = system.m((3e-9, 3e-9, 3e-9))
        assert np.linalg.norm(np.subtract(value, (0, Ms, 0))) < 1e-3

        value = system.m((0, 0, 0))
        assert np.linalg.norm(np.cross(value, (Ms, Ms, 0))) < 1e-3



import sys
import random
import pytest
import numpy as np
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestDMI:
    def setup(self):
        p1 = (-100e-9, 0, 0)
        p2 = (100e-9, 1e-9, 1e-9)
        self.region = df.Region(p1=p1, p2=p2)
        self.cell = (1e-9, 1e-9, 1e-9)
        self.subregions = {'r1': df.Region(p1=(-100e-9, 0, 0),
                                           p2=(0, 1e-9, 1e-9)),
                           'r2': df.Region(p1=(0, 0, 0),
                                           p2=(100e-9, 1e-9, 1e-9))}

    def random_m(self, pos):
        return [2*random.random()-1 for i in range(3)]

    def test_scalar(self):
        name = 'dmi_scalar'

        D = 1e-3
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.DMI(D=D, crystalclass='Cnv')

        mesh = df.Mesh(region=self.region, cell=self.cell)
        system.m = df.Field(mesh, dim=3, value=self.random_m, norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        # There are 4N cells in the mesh. Because of that the average should be
        # 0.
        assert np.linalg.norm(system.m.average) < 1

    def test_dict(self):
        name = 'dmi_dict'

        D = {'r1': 0, 'r2': 1e-3, 'default': 2e-3}
        Ms = 1e6

        system = mm.System(name=name)
        system.energy = mm.DMI(D=D, crystalclass='Cnv')

        mesh = df.Mesh(region=self.region, cell=self.cell,
                       subregions=self.subregions)
        system.m = df.Field(mesh, dim=3, value=self.random_m, norm=Ms)

        md = oc.MinDriver()
        md.drive(system)

        assert np.linalg.norm(system.m['r1'].average) > 1
        # There are 4N cells in the region with D!=0. Because of that
        # the average should be 0.
        assert np.linalg.norm(system.m['r2'].average) < 1

    @pytest.mark.skipif(sys.platform == 'win32',
                        reason=('Different crystalclasses are not '
                                'available on Windows.'))
    def test_crystalclass(self):
        name = 'dmi_crystalclass'

        D = 1e-3
        Ms = 1e6

        mesh = df.Mesh(region=self.region, cell=self.cell)

        for crystalclass in ['Cnv', 'T', 'O', 'D2d']:
            system = mm.System(name=name)
            system.energy = mm.DMI(D=D, crystalclass=crystalclass)

            system.m = df.Field(mesh, dim=3, value=self.random_m, norm=Ms)

            md = oc.MinDriver()
            md.drive(system)

            # There are 4N cells in the mesh. Because of that the
            # average should be 0.
            assert np.linalg.norm(system.m.average) < 1



import sys
import numbers
import oommfc as oc
import discretisedfield as df


def energy_script(container):
    mif = ''
    for term in container:
        mif += globals()[f'{term.name}_script'](term)

    return mif


def exchange_script(term):
    if isinstance(term.A, numbers.Real):
        mif = '# UniformExchange\n'
        mif += 'Specify Oxs_UniformExchange {\n'
        mif += f'  A {term.A}\n'
        mif += '}\n\n'

    elif isinstance(term.A, dict):
        if 'default' in term.A.keys():
            default_value = term.A['default']
        else:
            default_value = 0
        mif = '# Exchange6Ngbr\n'
        mif += 'Specify Oxs_Exchange6Ngbr {\n'
        mif += f'  default_A {default_value}\n'
        mif += '  atlas :main_atlas\n'
        mif += '  A {\n'
        for key, value in term.A.items():
            if key != 'default':
                if ':' in key:
                    region1, region2 = key.split(':')
                else:
                    region1, region2 = key, key
                mif += f'    {region1} {region2} {value}\n'
        mif += '  }\n'
        mif += '}\n\n'

    elif isinstance(term.A, df.Field):
        Amif, Aname = oc.scripts.setup_scalar_parameter(term.A, 'exchange_A')
        mif = Amif
        mif += '# ExchangePtwise\n'
        mif += 'Specify Oxs_ExchangePtwise {\n'
        mif += f'  A {Aname}\n'
        mif += '}\n\n'

    return mif


def zeeman_script(term):
    Hmif, Hname = oc.scripts.setup_vector_parameter(term.H, 'zeeman_H')

    mif = ''
    mif += Hmif
    mif += '# FixedZeeman\n'
    mif += 'Specify Oxs_FixedZeeman {\n'
    mif += f'  field {Hname}\n'
    mif += '}\n\n'

    return mif


def demag_script(term):
    mif = '# Demag\n'
    mif += 'Specify Oxs_Demag {\n'
    if hasattr(term, 'asymptotic_radius'):
        mif += f'  asymptotic_radius {term.asymptotic_radius}\n'
    mif += '}\n\n'

    return mif


def dmi_script(term):
    if term.crystalclass in ['T', 'O'] and sys.platform != 'win32':
        oxs = 'Oxs_DMI_T'
    elif term.crystalclass == 'D2d' and sys.platform != 'win32':
        oxs = 'Oxs_DMI_D2d'
    elif term.crystalclass == 'Cnv' and sys.platform != 'win32':
        oxs = 'Oxs_DMI_Cnv'
    # The following lines cannot be accessed on TravisCI, where coverage is
    # evaluated. Therefore, those lines are excluded from coverage.
    elif (term.crystalclass == 'Cnv' and
          sys.platform == 'win32'):  # pragma: no cover
        oxs = 'Oxs_DMExchange6Ngbr'  # pragma: no cover
    else:
        raise ValueError(f'The {term.crystalclass} crystal '
                         f'class is not supported on {sys.platform} '
                         'platform.')  # pragma: no cover

    mif = f'# DMI of crystallographic class {term.crystalclass}\n'
    mif += f'Specify {oxs} {{\n'

    if isinstance(term.D, numbers.Real):
        mif += f'  default_D {term.D}\n'
        mif += '  atlas :main_atlas\n'
        mif += '  D {\n'
        mif += f'    main main {term.D}\n'
        mif += '  }\n'
        mif += '}\n\n'

    elif isinstance(term.D, dict):
        if 'default' in term.D.keys():
            default_value = term.D['default']
        else:
            default_value = 0
        mif += f'  default_D {default_value}\n'
        mif += '  atlas :main_atlas\n'
        mif += '  D {\n'
        for key, value in term.D.items():
            if key != 'default':
                if ':' in key:
                    region1, region2 = key.split(':')
                else:
                    region1, region2 = key, key
                mif += f'    {region1} {region2} {value}\n'
        mif += '  }\n'
        mif += '}\n\n'

    return mif


def uniaxialanisotropy_script(term):
    kmif, kname = oc.scripts.setup_scalar_parameter(term.K, 'ua_K')
    umif, uname = oc.scripts.setup_vector_parameter(term.u, 'ua_u')

    mif = ''
    mif += kmif
    mif += umif
    mif += '# UniaxialAnisotropy\n'
    mif += 'Specify Oxs_UniaxialAnisotropy {\n'
    mif += f'  K1 {kname}\n'
    mif += f'  axis {uname}\n'
    mif += '}\n\n'

    return mif


def cubicanisotropy_script(term):
    kmif, kname = oc.scripts.setup_scalar_parameter(term.K, 'ca_K')
    u1mif, u1name = oc.scripts.setup_vector_parameter(term.u1, 'ca_u1')
    u2mif, u2name = oc.scripts.setup_vector_parameter(term.u2, 'ca_u2')

    mif = ''
    mif += kmif
    mif += u1mif
    mif += u2mif
    mif += '# CubicAnisotropy\n'
    mif += 'Specify Oxs_CubicAnisotropy {\n'
    mif += f'  K1 {kname}\n'
    mif += f'  axis1 {u1name}\n'
    mif += f'  axis2 {u2name}\n'
    mif += '}\n\n'

    return mif


def magnetoelastic_script(term):
    B1mif, B1name = oc.scripts.setup_scalar_parameter(term.B1, 'mel_B1')
    B2mif, B2name = oc.scripts.setup_scalar_parameter(term.B2, 'mel_B2')
    ediagmif, ediagname = oc.scripts.setup_vector_parameter(
        term.e_diag, 'mel_ediag')
    eoffdiagmif, eoffdiagname = oc.scripts.setup_vector_parameter(
        term.e_offdiag, 'mel_eoffdiag')

    mif = ''
    mif += B1mif
    mif += B2mif
    mif += ediagmif
    mif += eoffdiagmif
    mif += '# MagnetoElastic\n'
    mif += 'Specify YY_FixedMEL {\n'
    mif += f'  B1 {B1name}\n'
    mif += f'  B2 {B2name}\n'
    mif += f'  e_diag_field {ediagname}\n'
    mif += f'  e_offdiag_field {eoffdiagname}\n'
    mif += '}\n\n'

    return mif



import micromagneticmodel as mm


class RungeKuttaEvolver(mm.Evolver):
    """Runge-Kutta evolver.

    Only attributes in ``_allowed_attributes`` can be defined. For details on
    possible values for individual attributes and their default values, please
    refer to ``Oxs_RungeKuttaEvolver`` documentation
    (https://math.nist.gov/oommf/).

    Examples
    --------
    1. Defining evolver with a keyword argument.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.RungeKuttaEvolver(method='rk4')

    2. Passing an argument which is not allowed.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.RungeKuttaEvolver(myarg=3)
    Traceback (most recent call last):
       ...
    AttributeError: ...

    3. Getting the list of allowed attributes.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.RungeKuttaEvolver()
    >>> evolver._allowed_attributes
    [...]

    """
    _allowed_attributes = ['alpha',
                           'gamma_LL',
                           'gamma_G',
                           'do_precess',
                           'allow_signed_gamma',
                           'min_timestep',
                           'max_timestep',
                           'start_dm',
                           'start_dt',
                           'stage_start',
                           'error_rate',
                           'absolute_step_error',
                           'relative_step_error',
                           'energy_precision',
                           'min_step_headroom',
                           'max_step_headroom',
                           'reject_goal',
                           'method']



import micromagneticmodel as mm


class SpinTEvolver(mm.Evolver):
    """Zhang-Li evolver.

    Only attributes in ``_allowed_attributes`` can be defined. For details on
    possible values for individual attributes and their default values, please
    refer to ``Anv_SpinTEvolve`` documentation
    (https://www.zurich.ibm.com/st/nanomagnetism/spintevolve.html).

    Examples
    --------
    1. Defining evolver with a keyword argument.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.SpinTEvolver(method='rk4')

    2. Passing an argument which is not allowed.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.SpinTEvolver(myarg=3)
    Traceback (most recent call last):
       ...
    AttributeError: ...

    3. Getting the list of allowed attributes.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.SpinTEvolver()
    >>> evolver._allowed_attributes
    [...]

    """
    _allowed_attributes = ['alpha',
                           'gamma_LL',
                           'gamma_G',
                           'do_precess',
                           'u',
                           'beta',
                           'method']



import micromagneticmodel as mm


class CGEvolver(mm.Evolver):
    """Conjugate-Gradient evolver.

    Only attributes in ``_allowed_attributes`` can be defined. For details on
    possible values for individual attributes and their default values, please
    refer to ``Oxs_CGEvolver`` documentation (https://math.nist.gov/oommf/).

    Examples
    --------
    1. Defining evolver with a keyword argument.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.CGEvolver(method='Polak-Ribiere')

    2. Passing an argument which is not allowed.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.CGEvolver(myarg=3)
    Traceback (most recent call last):
       ...
    AttributeError: ...

    3. Getting the list of allowed attributes.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.CGEvolver()
    >>> evolver._allowed_attributes
    [...]

    """
    _allowed_attributes = ['gradient_reset_angle',
                           'gradient_reset_count',
                           'minimum_bracket_step',
                           'maximum_bracket_step',
                           'line_minimum_angle_precision',
                           'line_minimum_relwidth',
                           'energy_precision',
                           'method']



import oommfc as oc


def system_script(system, **kwargs):
    mif = '# MIF 2.2\n\n'
    # Output options
    mif += 'SetOptions {\n'
    mif += f'  basename {system.name}\n'
    mif += '  scalar_output_format %.12g\n'
    mif += '  scalar_field_output_format {text %#.15g}\n'
    mif += '  vector_field_output_format {text %#.15g}\n'
    mif += '}\n\n'

    # Mesh and energy scripts.
    mif += oc.scripts.mesh_script(system.m.mesh)
    mif += oc.scripts.energy_script(system.energy)

    # Magnetisation script.
    m0mif, _, _ = oc.scripts.setup_m0(system.m, 'm0')
    mif += m0mif

    return mif



import abc
import os
import glob
import json
import shutil
import tempfile
import datetime
import contextlib
import oommfc as oc
import ubermagtable as ut
import discretisedfield as df
import micromagneticmodel as mm


@contextlib.contextmanager
def _changedir(dirname):
    """Context manager for changing directory.

    """
    cwd = os.getcwd()
    os.chdir(dirname)
    try:
        yield
    finally:
        os.chdir(cwd)


class Driver(mm.Driver):
    @abc.abstractmethod
    def _checkargs(self, **kwargs):
        """Abstract method defined in a derived driver class.

        """
        pass  # pragma: no cover

    def _drive(self, system, basedirname, overwrite=False,
               compute=None, runner=None, **kwargs):
        """Convenience function, which allows to drive in different Python
        contexts.

        """
        # This method is implemented in the derived driver class. It raises
        # exception if any of the arguments are not valid.
        self._checkargs(**kwargs)

        # Generate directory.
        if compute is None:
            subdir = f'drive-{system.drive_number}'
        else:
            subdir = f'compute-{system.drive_number}'

        dirname = os.path.join(basedirname, system.name, subdir)
        # Check whether a directory already exists.
        if os.path.exists(dirname):
            if overwrite:
                oc.delete(system)
            else:
                msg = (f'Directory {dirname} already exists. To overwrite '
                       'it, pass overwrite=True to the drive method.')
                raise FileExistsError(msg)

        # Make a directory inside which OOMMF will be run.
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        # Generate the necessary filenames.
        miffilename = f'{system.name}.mif'
        jsonfilename = 'info.json'

        # Change directory to dirname
        with _changedir(dirname):
            # Generate and save mif file.
            mif = oc.scripts.system_script(system)
            mif += oc.scripts.driver_script(self, system, compute=compute,
                                            **kwargs)
            with open(miffilename, 'w') as miffile:
                miffile.write(mif)

            # Generate and save json info file.
            info = {}
            info['drive_number'] = system.drive_number
            info['date'] = datetime.datetime.now().strftime('%Y-%m-%d')
            info['time'] = datetime.datetime.now().strftime('%H:%M:%S')
            info['driver'] = self.__class__.__name__
            info['args'] = kwargs
            with open(jsonfilename, 'w') as jsonfile:
                jsonfile.write(json.dumps(info))

            # Run OOMMF.
            if runner is None:
                runner = oc.oommf.get_oommf_runner()
            runner.call(argstr=miffilename)

            # Update system's m and datatable attributes if the derivation of
            # E, Heff, or energy density was not asked.
            if compute is None:
                # Update system's magnetisation. An example .omf filename:
                # test_sample-Oxs_TimeDriver-Magnetization-01-0000008.omf
                omffiles = glob.iglob(f'{system.name}*.omf')
                lastomffile = sorted(omffiles)[-1]
                system.m.value = df.Field.fromfile(lastomffile)

                # Update system's datatable.
                system.table = ut.read(f'{system.name}.odt')

        # Increment drive_number independent of whether the files are saved
        # or not.
        if compute is None:
            system.drive_number += 1

    def drive(self, system, save=False, overwrite=False, compute=None,
              runner=None, **kwargs):
        """Drives the system in phase space.

        Takes ``micromagneticmodel.System`` and drives it in the phase space.
        If ``save=True``, the resulting files obtained from the OOMMF run are
        saved in the current directory (in ``system.name`` directory). If
        ``overwrite=True`` is passed, the directory with all previously created
        files (if exists) will be deleted before the system is run. To save a
        specific value during an OOMMF run ``Schedule...`` line can be passed
        using ``compute``. To specify the way OOMMF is run, an
        ``oommfc.oommf.OOMMFRunner`` can be passed using ``runner``.

        This method accepts any other arguments that could be required by the
        specific driver.

        Parameters
        ----------
        system : micromagneticmodel.System

          System to be driven.

        save : bool

            If ``True`` files created during an OOMMF run will be saved in the
            current directory. Defaults to ``False``.

        overwrite : bool

            If the directory from the previous drive already exists, it will be
            overwritten. Defaults to ``False``.

        compute : str

            ``Schedule...`` MIF line which can be added to the OOMMF file to
            save additional data. Defaults to ``None``.

        runner : oommfc.oommf.OOMMFRunner

            OOMMF Runner which is going to be used for running OOMMF. If
            ``None``, OOMMF runner will be found automatically. Defaults to
            ``None``.

        Examples
        --------
        1. Drive system using minimisation driver (``MinDriver``).

        >>> import micromagneticmodel as mm
        >>> import discretisedfield as df
        >>> import oommfc as oc
        ...
        >>> system = mm.System(name='my_cool_system')
        >>> system.energy = mm.Exchange(A=1e-12) + mm.Zeeman(H=(0, 0, 1e6))
        >>> mesh = df.Mesh(p1=(0, 0, 0), p2=(1e-9, 1e-9, 10e-9), n=(1, 1, 10))
        >>> system.m = df.Field(mesh, dim=3, value=(1, 1, 1), norm=1e6)
        ...
        >>> md = oc.MinDriver()
        >>> md.drive(system)
        Running OOMMF...

        2. Drive system using time driver (``TimeDriver``).

        >>> system.energy.zeeman.H = (0, 1e6, 0)
        ...
        >>> td = oc.TimeDriver()
        >>> td.drive(system, t=0.1e-9, n=10)
        Running OOMMF...

        """
        if save:
            self._drive(system=system, basedirname='', overwrite=overwrite,
                        compute=compute, runner=runner, **kwargs)
        else:
            with tempfile.TemporaryDirectory() as tmpdir:
                self._drive(system=system, basedirname=tmpdir,
                            overwrite=overwrite, compute=compute,
                            runner=runner, **kwargs)



import sys
import pytest
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


class TestCompute:
    def setup(self):
        name = 'compute_tests'
        p1 = (0, 0, 0)
        p2 = (10e-9, 2e-9, 2e-9)
        cell = (2e-9, 2e-9, 2e-9)
        region = df.Region(p1=p1, p2=p2)
        mesh = df.Mesh(region=region, cell=cell)

        self.system = mm.System(name=name)
        self.system.energy = (mm.Exchange(A=1e-12) +
                              mm.Demag() +
                              mm.Zeeman(H=(8e6, 0, 0)) +
                              mm.UniaxialAnisotropy(K=1e4, u=(0, 0, 1)) +
                              mm.CubicAnisotropy(K=1e3, u1=(1, 0, 0),
                                                 u2=(0, 1, 0)))

        self.system.m = df.Field(mesh, dim=3, value=(0, 0, 1), norm=8e6)

    def test_energy(self):
        for term in self.system.energy:
            assert isinstance(oc.compute(term.energy, self.system), float)
        assert isinstance(oc.compute(self.system.energy.energy, self.system),
                          float)

    def test_energy_density(self):
        for term in self.system.energy:
            assert isinstance(oc.compute(term.density, self.system), df.Field)
        assert isinstance(oc.compute(self.system.energy.density, self.system),
                          df.Field)

    def test_effective_field(self):
        for term in self.system.energy:
            assert isinstance(oc.compute(term.effective_field, self.system),
                              df.Field)
        assert isinstance(oc.compute(self.system.energy.effective_field,
                                     self.system), df.Field)

    def test_invalid_func(self):
        with pytest.raises(ValueError):
            val = oc.compute(self.system.energy.__len__, self.system)

    def test_dmi(self):
        if sys.platform != 'win32':
            self.system.energy += mm.DMI(D=5e-3, crystalclass='T')
            term = self.system.energy.dmi
            for crystalclass in ['T', 'Cnv', 'D2d']:
                term.crystalclass = crystalclass
                assert isinstance(oc.compute(term.energy, self.system),
                                  float)
                assert isinstance(oc.compute(term.density, self.system),
                                  df.Field)
                assert isinstance(oc.compute(term.effective_field,
                                             self.system), df.Field)



from .util import file_vector_field, vector_norm_scalar_field, \
    setup_scalar_parameter, setup_vector_parameter, \
    box_atlas, setup_m0
from .mesh import mesh_script
from .energy import energy_script
from .system import system_script
from .evolver import evolver_script
from .driver import driver_script



import os
import re
import glob
import json
import shutil
import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


def test_info_file():
    name = 'info_file'

    L = 30e-9   # (m)
    cell = (10e-9, 15e-9, 5e-9)  # (m)
    A = 1.3e-11  # (J/m)
    Ms = 8e5  # (A/m)
    H = (1e6, 0.0, 2e5)  # (A/m)
    gamma0 = 2.211e5  # (m/As)
    alpha = 0.02

    region = df.Region(p1=(0, 0, 0), p2=(L, L, L))
    mesh = df.Mesh(region=region, cell=cell)
    system = mm.System(name=name)
    system.energy = mm.Exchange(A=A) + mm.Zeeman(H=H)
    system.dynamics = mm.Precession(gamma0=gamma0) + mm.Damping(alpha=alpha)
    system.m = df.Field(mesh, dim=3, value=(0.0, 0.25, 0.1), norm=Ms)

    # First (0) drive
    td = oc.TimeDriver()
    td.drive(system, t=25e-12, n=10, save=True, overwrite=True)

    dirname = os.path.join(name, 'drive-0')
    infofile = os.path.join(dirname, 'info.json')
    assert os.path.exists(dirname)
    assert os.path.isfile(infofile)

    with open(infofile) as f:
        info = json.loads(f.read())
    assert 'drive_number' in info.keys()
    assert 'date' in info.keys()
    assert 'time' in info.keys()
    assert 'driver' in info.keys()
    assert 'args' in info.keys()

    assert info['drive_number'] == 0
    assert re.findall(r'\d{4}-\d{2}-\d{2}', info['date']) is not []
    assert re.findall(r'\d{2}:\d{2}-\d{2}', info['time']) is not []
    assert info['driver'] == 'TimeDriver'
    assert isinstance(info['args'], dict)
    assert info['args']['t'] == 25e-12
    assert info['args']['n'] == 10

    # Second (1) drive
    md = oc.MinDriver()
    md.drive(system, save=True)

    dirname = os.path.join(name, 'drive-1')
    infofile = os.path.join(dirname, 'info.json')
    assert os.path.exists(dirname)
    assert os.path.isfile(infofile)

    with open(infofile) as f:
        info = json.loads(f.read())
    assert 'drive_number' in info.keys()
    assert 'date' in info.keys()
    assert 'time' in info.keys()
    assert 'driver' in info.keys()
    assert 'args' in info.keys()

    assert info['drive_number'] == 1
    assert re.findall(r'\d{4}-\d{2}-\d{2}', info['date']) is not []
    assert re.findall(r'\d{2}:\d{2}-\d{2}', info['time']) is not []
    assert info['driver'] == 'MinDriver'
    assert isinstance(info['args'], dict)
    assert info['args'] == {}

    oc.delete(system)



import micromagneticmodel as mm


class EulerEvolver(mm.Evolver):
    """Euler evolver.

    Only attributes in ``_allowed_attributes`` can be defined. For details on
    possible values for individual attributes and their default values, please
    refer to ``Oxs_EulerEvolver`` documentation (https://math.nist.gov/oommf/).

    Examples
    --------
    1. Defining evolver with a keyword argument.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.EulerEvolver(min_timestep=0)

    2. Passing an argument which is not allowed.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.EulerEvolver(myarg=3)
    Traceback (most recent call last):
       ...
    AttributeError: ...

    3. Getting the list of allowed attributes.

    >>> import oommfc as oc
    ...
    >>> evolver = oc.EulerEvolver()
    >>> evolver._allowed_attributes
    [...]

    """
    _allowed_attributes = ['alpha',
                           'gamma_LL',
                           'gamma_G',
                           'do_precess',
                           'min_timestep',
                           'max_timestep',
                           'start_dm',
                           'error_rate',
                           'absolute_step_error',
                           'relative_step_error',
                           'step_headroom']



import os
import pytest
import oommfc as oc
import micromagneticmodel as mm


def test_save_delete():
    system = mm.examples.macrospin()

    td = oc.TimeDriver()
    td.drive(system, t=1e-12, n=5, save=True, overwrite=True)

    assert os.path.exists(os.path.join(system.name, 'drive-0'))

    with pytest.raises(FileExistsError):
        system.drive_number = 0
        td.drive(system, t=1e-12, n=5, save=True)

    assert os.path.exists(os.path.join(system.name, 'drive-0'))
    assert not os.path.exists(os.path.join(system.name, 'drive-1'))

    system.drive_number = 0
    td.drive(system, t=1e-12, n=5, save=False)

    assert os.path.exists(os.path.join(system.name, 'drive-0'))
    assert not os.path.exists(os.path.join(system.name, 'drive-1'))

    td.drive(system, t=1e-12, n=5, save=True, overwrite=True)

    assert os.path.exists(os.path.join(system.name, 'drive-0'))
    assert os.path.exists(os.path.join(system.name, 'drive-1'))

    oc.delete(system)

    assert not os.path.exists(system.name)

    with pytest.raises(FileNotFoundError):
        oc.delete(system)



import os
import sys
import pytest
import shutil
import oommfc as oc
import oommfc.oommf as oo
import micromagneticmodel as mm


def check_runner(runner):
    # Testing OOMMF on a mif file to make it independent of oommfc.
    dirname = os.path.join(os.path.dirname(__file__), 'test_files')
    os.chdir(dirname)
    argstr = 'test_oommf.mif'
    res = runner.call(argstr)
    version = runner.version()
    platform = runner.platform()

    assert isinstance(version, str)
    assert len(version) > 0
    assert isinstance(platform, str)
    assert len(platform) > 0
    assert res.returncode == 0

    # Cleanup created files.
    for f in os.listdir(dirname):
        if f.endswith('.odt'):
            os.remove(os.path.join(dirname, f))
        elif f.endswith('.omf') and f.startswith('test_oommf-Oxs'):
            os.remove(os.path.join(dirname, f))


@pytest.mark.travis
def test_tcl_oommf_runner():
    # TclOOMMFRunner runs when OOMMFTCL environment variable is set. On
    # TravisCI OOMMFTCL environment variable is set inside ubermag/oommf docker
    # image.
    oommf_tcl = os.environ.get('OOMMFTCL', None)
    runner = oo.TclOOMMFRunner(oommf_tcl)
    check_runner(runner)
    assert isinstance(runner.errors(), str)


@pytest.mark.travis
def test_exe_oommf_runner():
    # ExeOOMMFRunner runs when callable OOMMF exists ('oommf'). On TravisCI
    # oommf is an executable inside ubermag/oommf docker image.
    oommf_exe = 'oommf'
    runner = oo.ExeOOMMFRunner(oommf_exe)
    check_runner(runner)
    with pytest.raises(EnvironmentError):
        # On travis oommf compiled from source is used.
        errors = runner.errors()


@pytest.mark.docker
def test_docker_oommf_runner():
    # DockerOOMMFRunner runs when docker is installed. This test does not run
    # on host or TravisCI. It can be run using make test-docker on host if
    # docker is installed.
    docker_exe = 'docker'
    image = 'ubermag/oommf'
    runner = oo.DockerOOMMFRunner(docker_exe, image)
    check_runner(runner)

    # An additional check of getting OOMMF runner when docker is installed.
    runner = oo.get_oommf_runner(use_cache=False,
                                 envvar='wrong_name',
                                 oommf_exe='wrong_name',
                                 docker_exe='docker')
    assert isinstance(runner, oo.DockerOOMMFRunner)
    check_runner(runner)

    with pytest.raises(EnvironmentError):
        errors = runner.errors()


@pytest.mark.travis
def test_get_oommf_runner():
    # TclOOMMFRunner
    oommf_runner = oo.get_oommf_runner(use_cache=False,
                                       envvar='OOMMFTCL',
                                       oommf_exe='wrong_name',
                                       docker_exe='wrong_name')
    assert isinstance(oommf_runner, oo.TclOOMMFRunner)
    check_runner(oommf_runner)

    # ExeOOMMFRunner
    oommf_runner = oo.get_oommf_runner(use_cache=False,
                                       envvar='wrong_name',
                                       oommf_exe='oommf',
                                       docker_exe='wrong_name')
    assert isinstance(oommf_runner, oo.ExeOOMMFRunner)
    check_runner(oommf_runner)

    # OOMMF cannot be found on the system.
    with pytest.raises(EnvironmentError):
        oommf_runner = oo.get_oommf_runner(use_cache=False,
                                           envvar='wrong_name',
                                           oommf_exe='wrong_name',
                                           docker_exe='wrong_name')

    # Test cached OOMMFRunner.
    oommf_runner = oo.get_oommf_runner(use_cache=False,
                                       envvar='wrong_name',
                                       oommf_exe='oommf',
                                       docker_exe='wrong_name')
    assert isinstance(oommf_runner, oo.ExeOOMMFRunner)
    check_runner(oommf_runner)

    oommf_runner = oo.get_oommf_runner(use_cache=True)
    assert isinstance(oommf_runner, oo.ExeOOMMFRunner)
    check_runner(oommf_runner)

    oommf_runner = oo.get_oommf_runner(use_cache=True,
                                       envvar='OOMMFTCL',
                                       oommf_exe='wrong_name',
                                       docker_exe='wrong_name')
    assert isinstance(oommf_runner, oo.ExeOOMMFRunner)
    check_runner(oommf_runner)


def test_get_oommf_runner():
    # This is a shorter version of the previous test for testing on host.
    oommf_runner = oo.get_oommf_runner(use_cache=False)
    assert isinstance(oommf_runner, oo.OOMMFRunner)
    check_runner(oommf_runner)


def test_status():
    assert oo.status() == 0


def test_overhead():
    assert isinstance(oo.overhead(), float)


def test_runtimeerror():
    oommf_runner = oo.get_oommf_runner(use_cache=False)
    with pytest.raises(RuntimeError):
        oommf_runner.call('+wrong_argstr')


@pytest.mark.travis
def test_choose_runner():
    system = mm.examples.macrospin()

    md = oc.MinDriver()
    runner = oc.oommf.TclOOMMFRunner(oommf_tcl=os.environ.get('OOMMFTCL',
                                                              None))
    md.drive(system, runner=runner)

    runner = oc.oommf.ExeOOMMFRunner(oommf_exe='oommf')
    md.drive(system, runner=runner)



import oommfc as oc
import discretisedfield as df
import micromagneticmodel as mm


def test_stdprob4():
    name = 'stdprob4'

    L, d, th = 500e-9, 125e-9, 3e-9   # (m)
    cell = (5e-9, 5e-9, 3e-9)  # (m)
    p1 = (0, 0, 0)
    p2 = (L, d, th)
    region = df.Region(p1=p1, p2=p2)
    mesh = df.Mesh(region=region, cell=cell)

    Ms = 8e5  # (A/m)
    A = 1.3e-11  # (J/m)

    system = mm.System(name=name)
    system.energy = mm.Exchange(A=A) + mm.Demag()

    gamma0 = 2.211e5  # (m/As)
    alpha = 0.02
    system.dynamics = mm.Precession(gamma0=gamma0) + mm.Damping(alpha=alpha)

    system.m = df.Field(mesh, dim=3, value=(1, 0.25, 0.1), norm=Ms)

    md = oc.MinDriver()
    md.drive(system)  # updates system.m in-place

    H = (-24.6e-3/mm.consts.mu0, 4.3e-3/mm.consts.mu0, 0)
    system.energy += mm.Zeeman(H=H)

    td = oc.TimeDriver()
    td.drive(system, t=1e-9, n=200)

    t = system.table['t'].values
    my = system.table['my'].values

    assert abs(min(t) - 5e-12) < 1e-20
    assert abs(max(t) - 1e-9) < 1e-20

    # Eye-norm test.
    assert 0.7 < max(my) < 0.8
    assert -0.5 < min(my) < -0.4
